@@ -108,7 +108,7 @@ function emitPromiseRejectionWarnings() {
     }
   }
 
-  let hadListeners = false;
+  let maybeScheduledTicks = false;
   let len = pendingUnhandledRejections.length;
   while (len--) {
     const promise = pendingUnhandledRejections.shift();
@@ -118,10 +118,9 @@ function emitPromiseRejectionWarnings() {
       const { reason, uid } = promiseInfo;
       if (!process.emit('unhandledRejection', reason, promise)) {
         emitWarning(uid, reason);
-      } else {
-        hadListeners = true;
       }
+      maybeScheduledTicks = true;
     }
   }
-  return hadListeners || pendingUnhandledRejections.length !== 0;
+  return maybeScheduledTicks || pendingUnhandledRejections.length !== 0;
 }
