@@ -186,20 +186,20 @@ inline void Environment::AsyncHooks::clear_async_id_stack() {
 inline Environment::AsyncHooks::DefaultTriggerAsyncIdScope
   ::DefaultTriggerAsyncIdScope(Environment* env,
                                double default_trigger_async_id)
-    : async_id_fields_ref_(env->async_hooks()->async_id_fields()) {
+    : async_hooks_(env->async_hooks()) {
   if (env->async_hooks()->fields()[AsyncHooks::kCheck] > 0) {
     CHECK_GE(default_trigger_async_id, 0);
   }
 
   old_default_trigger_async_id_ =
-    async_id_fields_ref_[AsyncHooks::kDefaultTriggerAsyncId];
-  async_id_fields_ref_[AsyncHooks::kDefaultTriggerAsyncId] =
+    async_hooks_->async_id_fields()[AsyncHooks::kDefaultTriggerAsyncId];
+  async_hooks_->async_id_fields()[AsyncHooks::kDefaultTriggerAsyncId] =
     default_trigger_async_id;
 }
 
 inline Environment::AsyncHooks::DefaultTriggerAsyncIdScope
   ::~DefaultTriggerAsyncIdScope() {
-  async_id_fields_ref_[AsyncHooks::kDefaultTriggerAsyncId] =
+  async_hooks_->async_id_fields()[AsyncHooks::kDefaultTriggerAsyncId] =
     old_default_trigger_async_id_;
 }
 
@@ -441,6 +441,16 @@ Environment::should_abort_on_uncaught_toggle() {
   return should_abort_on_uncaught_toggle_;
 }
 
+inline uint32_t Environment::get_next_module_id() {
+  return module_id_counter_++;
+}
+inline uint32_t Environment::get_next_script_id() {
+  return script_id_counter_++;
+}
+inline uint32_t Environment::get_next_function_id() {
+  return function_id_counter_++;
+}
+
 Environment::ShouldNotAbortOnUncaughtScope::ShouldNotAbortOnUncaughtScope(
     Environment* env)
     : env_(env) {
@@ -583,8 +593,7 @@ void Environment::CreateImmediate(native_immediate_callback cb,
   native_immediate_callbacks_.push_back({
     cb,
     data,
-    std::unique_ptr<Persistent<v8::Object>>(obj.IsEmpty() ?
-        nullptr : new Persistent<v8::Object>(isolate_, obj)),
+    v8::Global<v8::Object>(isolate_, obj),
     ref
   });
   immediate_info()->count_inc(1);
