@@ -1,15 +1,42 @@
+// Copyright Joyent, Inc. and other Node contributors.
+//
+// Permission is hereby granted, free of charge, to any person obtaining a
+// copy of this software and associated documentation files (the
+// "Software"), to deal in the Software without restriction, including
+// without limitation the rights to use, copy, modify, merge, publish,
+// distribute, sublicense, and/or sell copies of the Software, and to permit
+// persons to whom the Software is furnished to do so, subject to the
+// following conditions:
+//
+// The above copyright notice and this permission notice shall be included
+// in all copies or substantial portions of the Software.
+//
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
+// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
+// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
+// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
+// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
+// USE OR OTHER DEALINGS IN THE SOFTWARE.
+
 'use strict';
 const common = require('../common');
+const fixtures = require('../common/fixtures');
+const tmpdir = require('../common/tmpdir');
+
+if (!common.isMainThread)
+  common.skip('process.chdir is not available in Workers');
+
 const assert = require('assert');
 const fs = require('fs');
 const path = require('path');
-const exec = require('child_process').exec;
-let async_completed = 0, async_expected = 0;
+let async_completed = 0;
+let async_expected = 0;
 const unlink = [];
-let skipSymlinks = false;
-const tmpDir = common.tmpDir;
+const skipSymlinks = !common.canCreateSymLink();
+const tmpDir = tmpdir.path;
 
-common.refreshTmpDir();
+tmpdir.refresh();
 
 let root = '/';
 let assertEqualPath = assert.strictEqual;
@@ -20,25 +47,9 @@ if (common.isWindows) {
     assert
       .strictEqual(path_left.toLowerCase(), path_right.toLowerCase(), message);
   };
-
-  // On Windows, creating symlinks requires admin privileges.
-  // We'll only try to run symlink test if we have enough privileges.
-  try {
-    exec('whoami /priv', function(err, o) {
-      if (err || !o.includes('SeCreateSymbolicLinkPrivilege')) {
-        skipSymlinks = true;
-      }
-      runTest();
-    });
-  } catch (er) {
-    // better safe than sorry
-    skipSymlinks = true;
-    process.nextTick(runTest);
-  }
-} else {
-  process.nextTick(runTest);
 }
 
+process.nextTick(runTest);
 
 function tmp(p) {
   return path.join(tmpDir, p);
@@ -70,38 +81,38 @@ function asynctest(testBlock, args, callback, assertBlock) {
 }
 
 // sub-tests:
-function test_simple_error_callback(cb) {
-  fs.realpath('/this/path/does/not/exist', common.mustCall(function(err, s) {
+function test_simple_error_callback(realpath, realpathSync, cb) {
+  realpath('/this/path/does/not/exist', common.mustCall(function(err, s) {
     assert(err);
     assert(!s);
     cb();
   }));
 }
 
-function test_simple_relative_symlink(callback) {
+function test_simple_relative_symlink(realpath, realpathSync, callback) {
   console.log('test_simple_relative_symlink');
   if (skipSymlinks) {
-    common.skip('symlink test (no privs)');
-    return runNextTest();
+    common.printSkipMessage('symlink test (no privs)');
+    return callback();
   }
   const entry = `${tmpDir}/symlink`;
   const expected = `${tmpDir}/cycles/root.js`;
   [
-    [entry, `../${common.tmpDirName}/cycles/root.js`]
+    [entry, `../${path.basename(tmpDir)}/cycles/root.js`]
   ].forEach(function(t) {
-    try { fs.unlinkSync(t[0]); } catch (e) {}
+    try { fs.unlinkSync(t[0]); } catch {}
     console.log('fs.symlinkSync(%j, %j, %j)', t[1], t[0], 'file');
     fs.symlinkSync(t[1], t[0], 'file');
     unlink.push(t[0]);
   });
-  const result = fs.realpathSync(entry);
+  const result = realpathSync(entry);
   assertEqualPath(result, path.resolve(expected));
-  asynctest(fs.realpath, [entry], callback, function(err, result) {
+  asynctest(realpath, [entry], callback, function(err, result) {
     assertEqualPath(result, path.resolve(expected));
   });
 }
 
-function test_simple_absolute_symlink(callback) {
+function test_simple_absolute_symlink(realpath, realpathSync, callback) {
   console.log('test_simple_absolute_symlink');
 
   // this one should still run, even if skipSymlinks is set,
@@ -111,85 +122,85 @@ function test_simple_absolute_symlink(callback) {
   console.log('using type=%s', type);
 
   const entry = `${tmpAbsDir}/symlink`;
-  const expected = `${common.fixturesDir}/nested-index/one`;
+  const expected = fixtures.path('nested-index', 'one');
   [
     [entry, expected]
   ].forEach(function(t) {
-    try { fs.unlinkSync(t[0]); } catch (e) {}
+    try { fs.unlinkSync(t[0]); } catch {}
     console.error('fs.symlinkSync(%j, %j, %j)', t[1], t[0], type);
     fs.symlinkSync(t[1], t[0], type);
     unlink.push(t[0]);
   });
-  const result = fs.realpathSync(entry);
+  const result = realpathSync(entry);
   assertEqualPath(result, path.resolve(expected));
-  asynctest(fs.realpath, [entry], callback, function(err, result) {
+  asynctest(realpath, [entry], callback, function(err, result) {
     assertEqualPath(result, path.resolve(expected));
   });
 }
 
-function test_deep_relative_file_symlink(callback) {
+function test_deep_relative_file_symlink(realpath, realpathSync, callback) {
   console.log('test_deep_relative_file_symlink');
   if (skipSymlinks) {
-    common.skip('symlink test (no privs)');
-    return runNextTest();
+    common.printSkipMessage('symlink test (no privs)');
+    return callback();
   }
 
-  const expected = path.join(common.fixturesDir, 'cycles', 'root.js');
+  const expected = fixtures.path('cycles', 'root.js');
   const linkData1 = path
                       .relative(path.join(targetsAbsDir, 'nested-index', 'one'),
                                 expected);
   const linkPath1 = path.join(targetsAbsDir,
                               'nested-index', 'one', 'symlink1.js');
-  try { fs.unlinkSync(linkPath1); } catch (e) {}
+  try { fs.unlinkSync(linkPath1); } catch {}
   fs.symlinkSync(linkData1, linkPath1, 'file');
 
   const linkData2 = '../one/symlink1.js';
   const entry = path.join(targetsAbsDir,
                           'nested-index', 'two', 'symlink1-b.js');
-  try { fs.unlinkSync(entry); } catch (e) {}
+  try { fs.unlinkSync(entry); } catch {}
   fs.symlinkSync(linkData2, entry, 'file');
   unlink.push(linkPath1);
   unlink.push(entry);
 
-  assertEqualPath(fs.realpathSync(entry), path.resolve(expected));
-  asynctest(fs.realpath, [entry], callback, function(err, result) {
+  assertEqualPath(realpathSync(entry), path.resolve(expected));
+  asynctest(realpath, [entry], callback, function(err, result) {
     assertEqualPath(result, path.resolve(expected));
   });
 }
 
-function test_deep_relative_dir_symlink(callback) {
+function test_deep_relative_dir_symlink(realpath, realpathSync, callback) {
   console.log('test_deep_relative_dir_symlink');
   if (skipSymlinks) {
-    common.skip('symlink test (no privs)');
-    return runNextTest();
+    common.printSkipMessage('symlink test (no privs)');
+    return callback();
   }
-  const expected = path.join(common.fixturesDir, 'cycles', 'folder');
+  const expected = fixtures.path('cycles', 'folder');
   const path1b = path.join(targetsAbsDir, 'nested-index', 'one');
   const linkPath1b = path.join(path1b, 'symlink1-dir');
   const linkData1b = path.relative(path1b, expected);
-  try { fs.unlinkSync(linkPath1b); } catch (e) {}
+  try { fs.unlinkSync(linkPath1b); } catch {}
   fs.symlinkSync(linkData1b, linkPath1b, 'dir');
 
   const linkData2b = '../one/symlink1-dir';
   const entry = path.join(targetsAbsDir,
                           'nested-index', 'two', 'symlink12-dir');
-  try { fs.unlinkSync(entry); } catch (e) {}
+  try { fs.unlinkSync(entry); } catch {}
   fs.symlinkSync(linkData2b, entry, 'dir');
   unlink.push(linkPath1b);
   unlink.push(entry);
 
-  assertEqualPath(fs.realpathSync(entry), path.resolve(expected));
+  assertEqualPath(realpathSync(entry), path.resolve(expected));
 
-  asynctest(fs.realpath, [entry], callback, function(err, result) {
+  asynctest(realpath, [entry], callback, function(err, result) {
     assertEqualPath(result, path.resolve(expected));
   });
 }
 
-function test_cyclic_link_protection(callback) {
+function test_cyclic_link_protection(realpath, realpathSync, callback) {
   console.log('test_cyclic_link_protection');
   if (skipSymlinks) {
-    common.skip('symlink test (no privs)');
-    return runNextTest();
+    common.printSkipMessage('symlink test (no privs)');
+    return callback();
   }
   const entry = path.join(tmpDir, '/cycles/realpath-3a');
   [
@@ -197,47 +208,47 @@ function test_cyclic_link_protection(callback) {
     [path.join(tmpDir, '/cycles/realpath-3b'), '../cycles/realpath-3c'],
     [path.join(tmpDir, '/cycles/realpath-3c'), '../cycles/realpath-3a']
   ].forEach(function(t) {
-    try { fs.unlinkSync(t[0]); } catch (e) {}
+    try { fs.unlinkSync(t[0]); } catch {}
     fs.symlinkSync(t[1], t[0], 'dir');
     unlink.push(t[0]);
   });
-  assert.throws(() => {
-    fs.realpathSync(entry);
-  }, /^Error: ELOOP: too many symbolic links encountered, stat /);
+  common.expectsError(() => {
+    realpathSync(entry);
+  }, { code: 'ELOOP', type: Error });
   asynctest(
-    fs.realpath, [entry], callback, common.mustCall(function(err, result) {
+    realpath, [entry], callback, common.mustCall(function(err, result) {
       assert.strictEqual(err.path, entry);
       assert.strictEqual(result, undefined);
       return true;
     }));
 }
 
-function test_cyclic_link_overprotection(callback) {
+function test_cyclic_link_overprotection(realpath, realpathSync, callback) {
   console.log('test_cyclic_link_overprotection');
   if (skipSymlinks) {
-    common.skip('symlink test (no privs)');
-    return runNextTest();
+    common.printSkipMessage('symlink test (no privs)');
+    return callback();
   }
   const cycles = `${tmpDir}/cycles`;
-  const expected = fs.realpathSync(cycles);
+  const expected = realpathSync(cycles);
   const folder = `${cycles}/folder`;
   const link = `${folder}/cycles`;
   let testPath = cycles;
   testPath += '/folder/cycles'.repeat(10);
-  try { fs.unlinkSync(link); } catch (ex) {}
+  try { fs.unlinkSync(link); } catch {}
   fs.symlinkSync(cycles, link, 'dir');
   unlink.push(link);
-  assertEqualPath(fs.realpathSync(testPath), path.resolve(expected));
-  asynctest(fs.realpath, [testPath], callback, function(er, res) {
+  assertEqualPath(realpathSync(testPath), path.resolve(expected));
+  asynctest(realpath, [testPath], callback, function(er, res) {
     assertEqualPath(res, path.resolve(expected));
   });
 }
 
-function test_relative_input_cwd(callback) {
+function test_relative_input_cwd(realpath, realpathSync, callback) {
   console.log('test_relative_input_cwd');
   if (skipSymlinks) {
-    common.skip('symlink test (no privs)');
-    return runNextTest();
+    common.printSkipMessage('symlink test (no privs)');
+    return callback();
   }
 
   // we need to calculate the relative path to the tmp dir from cwd
@@ -252,7 +263,7 @@ function test_relative_input_cwd(callback) {
   ].forEach(function(t) {
     const fn = t[0];
     console.error('fn=%j', fn);
-    try { fs.unlinkSync(fn); } catch (e) {}
+    try { fs.unlinkSync(fn); } catch {}
     const b = path.basename(t[1]);
     const type = (b === 'root.js' ? 'file' : 'dir');
     console.log('fs.symlinkSync(%j, %j, %j)', t[1], fn, type);
@@ -262,21 +273,21 @@ function test_relative_input_cwd(callback) {
 
   const origcwd = process.cwd();
   process.chdir(entrydir);
-  assertEqualPath(fs.realpathSync(entry), path.resolve(expected));
-  asynctest(fs.realpath, [entry], callback, function(err, result) {
+  assertEqualPath(realpathSync(entry), path.resolve(expected));
+  asynctest(realpath, [entry], callback, function(err, result) {
     process.chdir(origcwd);
     assertEqualPath(result, path.resolve(expected));
     return true;
   });
 }
 
-function test_deep_symlink_mix(callback) {
+function test_deep_symlink_mix(realpath, realpathSync, callback) {
   console.log('test_deep_symlink_mix');
   if (common.isWindows) {
     // This one is a mix of files and directories, and it's quite tricky
     // to get the file/dir links sorted out correctly.
-    common.skip('symlink test (no privs)');
-    return runNextTest();
+    common.printSkipMessage('symlink test (no privs)');
+    return callback();
   }
 
   /*
@@ -291,8 +302,8 @@ function test_deep_symlink_mix(callback) {
   $tmpDir/targets/cycles/root.js (hard)
   */
   const entry = tmp('node-test-realpath-f1');
-  try { fs.unlinkSync(tmp('node-test-realpath-d2/foo')); } catch (e) {}
-  try { fs.rmdirSync(tmp('node-test-realpath-d2')); } catch (e) {}
+  try { fs.unlinkSync(tmp('node-test-realpath-d2/foo')); } catch {}
+  try { fs.rmdirSync(tmp('node-test-realpath-d2')); } catch {}
   fs.mkdirSync(tmp('node-test-realpath-d2'), 0o700);
   try {
     [
@@ -307,7 +318,7 @@ function test_deep_symlink_mix(callback) {
       [`${targetsAbsDir}/nested-index/two/realpath-c`,
        `${tmpDir}/cycles/root.js`]
     ].forEach(function(t) {
-      try { fs.unlinkSync(t[0]); } catch (e) {}
+      try { fs.unlinkSync(t[0]); } catch {}
       fs.symlinkSync(t[1], t[0]);
       unlink.push(t[0]);
     });
@@ -315,22 +326,22 @@ function test_deep_symlink_mix(callback) {
     unlink.push(tmp('node-test-realpath-d2'));
   }
   const expected = `${tmpAbsDir}/cycles/root.js`;
-  assertEqualPath(fs.realpathSync(entry), path.resolve(expected));
-  asynctest(fs.realpath, [entry], callback, function(err, result) {
+  assertEqualPath(realpathSync(entry), path.resolve(expected));
+  asynctest(realpath, [entry], callback, function(err, result) {
     assertEqualPath(result, path.resolve(expected));
     return true;
   });
 }
 
-function test_non_symlinks(callback) {
+function test_non_symlinks(realpath, realpathSync, callback) {
   console.log('test_non_symlinks');
   const entrydir = path.dirname(tmpAbsDir);
   const entry = `${tmpAbsDir.substr(entrydir.length + 1)}/cycles/root.js`;
   const expected = `${tmpAbsDir}/cycles/root.js`;
   const origcwd = process.cwd();
   process.chdir(entrydir);
-  assertEqualPath(fs.realpathSync(entry), path.resolve(expected));
-  asynctest(fs.realpath, [entry], callback, function(err, result) {
+  assertEqualPath(realpathSync(entry), path.resolve(expected));
+  asynctest(realpath, [entry], callback, function(err, result) {
     process.chdir(origcwd);
     assertEqualPath(result, path.resolve(expected));
     return true;
@@ -338,19 +349,21 @@ function test_non_symlinks(callback) {
 }
 
 const upone = path.join(process.cwd(), '..');
-function test_escape_cwd(cb) {
+function test_escape_cwd(realpath, realpathSync, cb) {
   console.log('test_escape_cwd');
-  asynctest(fs.realpath, ['..'], cb, function(er, uponeActual) {
+  asynctest(realpath, ['..'], cb, function(er, uponeActual) {
     assertEqualPath(
       upone, uponeActual,
       `realpath("..") expected: ${path.resolve(upone)} actual:${uponeActual}`);
   });
 }
-const uponeActual = fs.realpathSync('..');
-assertEqualPath(
-  upone, uponeActual,
-  `realpathSync("..") expected: ${path.resolve(upone)} actual:${uponeActual}`);
 
+function test_upone_actual(realpath, realpathSync, cb) {
+  console.log('test_upone_actual');
+  const uponeActual = realpathSync('..');
+  assertEqualPath(upone, uponeActual);
+  cb();
+}
 
 // going up with .. multiple times
 // .
@@ -359,23 +372,14 @@ assertEqualPath(
 //     |   `-- e -> ..
 //     `-- d -> ..
 // realpath(a/b/e/d/a/b/e/d/a) ==> a
-function test_up_multiple(cb) {
+function test_up_multiple(realpath, realpathSync, cb) {
   console.error('test_up_multiple');
   if (skipSymlinks) {
-    common.skip('symlink test (no privs)');
-    return runNextTest();
+    common.printSkipMessage('symlink test (no privs)');
+    return cb();
   }
-  function cleanup() {
-    ['a/b',
-     'a'
-    ].forEach(function(folder) {
-      try { fs.rmdirSync(tmp(folder)); } catch (ex) {}
-    });
-  }
-  function setup() {
-    cleanup();
-  }
-  setup();
+  const tmpdir = require('../common/tmpdir');
+  tmpdir.refresh();
   fs.mkdirSync(tmp('a'), 0o755);
   fs.mkdirSync(tmp('a/b'), 0o755);
   fs.symlinkSync('..', tmp('a/d'), 'dir');
@@ -389,16 +393,15 @@ function test_up_multiple(cb) {
   const abedabeda = tmp('abedabeda'.split('').join('/'));
   const abedabeda_real = tmp('a');
 
-  assertEqualPath(fs.realpathSync(abedabeda), abedabeda_real);
-  assertEqualPath(fs.realpathSync(abedabed), abedabed_real);
-  fs.realpath(abedabeda, function(er, real) {
-    if (er) throw er;
+  assertEqualPath(realpathSync(abedabeda), abedabeda_real);
+  assertEqualPath(realpathSync(abedabed), abedabed_real);
+  realpath(abedabeda, function(er, real) {
+    assert.ifError(er);
     assertEqualPath(abedabeda_real, real);
-    fs.realpath(abedabed, function(er, real) {
-      if (er) throw er;
+    realpath(abedabed, function(er, real) {
+      assert.ifError(er);
       assertEqualPath(abedabed_real, real);
       cb();
-      cleanup();
     });
   });
 }
@@ -412,7 +415,7 @@ function test_up_multiple(cb) {
 //     |       `-- x.txt
 //     `-- link -> /tmp/node-test-realpath-abs-kids/a/b/
 // realpath(root+'/a/link/c/x.txt') ==> root+'/a/b/c/x.txt'
-function test_abs_with_kids(cb) {
+function test_abs_with_kids(realpath, realpathSync, cb) {
   console.log('test_abs_with_kids');
 
   // this one should still run, even if skipSymlinks is set,
@@ -426,14 +429,14 @@ function test_abs_with_kids(cb) {
     ['/a/b/c/x.txt',
      '/a/link'
     ].forEach(function(file) {
-      try { fs.unlinkSync(root + file); } catch (ex) {}
+      try { fs.unlinkSync(root + file); } catch {}
     });
     ['/a/b/c',
      '/a/b',
      '/a',
      ''
     ].forEach(function(folder) {
-      try { fs.rmdirSync(root + folder); } catch (ex) {}
+      try { fs.rmdirSync(root + folder); } catch {}
     });
   }
   function setup() {
@@ -452,16 +455,25 @@ function test_abs_with_kids(cb) {
   setup();
   const linkPath = `${root}/a/link/c/x.txt`;
   const expectPath = `${root}/a/b/c/x.txt`;
-  const actual = fs.realpathSync(linkPath);
+  const actual = realpathSync(linkPath);
   // console.log({link:linkPath,expect:expectPath,actual:actual},'sync');
   assertEqualPath(actual, path.resolve(expectPath));
-  asynctest(fs.realpath, [linkPath], cb, function(er, actual) {
+  asynctest(realpath, [linkPath], cb, function(er, actual) {
     // console.log({link:linkPath,expect:expectPath,actual:actual},'async');
     assertEqualPath(actual, path.resolve(expectPath));
     cleanup();
   });
 }
 
+function test_root(realpath, realpathSync, cb) {
+  assertEqualPath(root, realpathSync('/'));
+  realpath('/', function(err, result) {
+    assert.ifError(err);
+    assertEqualPath(root, result);
+    cb();
+  });
+}
+
 // ----------------------------------------------------------------------------
 
 const tests = [
@@ -476,29 +488,29 @@ const tests = [
   test_deep_symlink_mix,
   test_non_symlinks,
   test_escape_cwd,
+  test_upone_actual,
   test_abs_with_kids,
-  test_up_multiple
+  test_up_multiple,
+  test_root,
 ];
 const numtests = tests.length;
 let testsRun = 0;
 function runNextTest(err) {
-  if (err) throw err;
+  assert.ifError(err);
   const test = tests.shift();
   if (!test) {
     return console.log(`${numtests} subtests completed OK for fs.realpath`);
   }
   testsRun++;
-  test(runNextTest);
+  test(fs.realpath, fs.realpathSync, common.mustCall((err) => {
+    assert.ifError(err);
+    testsRun++;
+    test(fs.realpath.native,
+         fs.realpathSync.native,
+         common.mustCall(runNextTest));
+  }));
 }
 
-
-assertEqualPath(root, fs.realpathSync('/'));
-fs.realpath('/', function(err, result) {
-  assert.ifError(err);
-  assertEqualPath(root, result);
-});
-
-
 function runTest() {
   const tmpDirs = ['cycles', 'cycles/folder'];
   tmpDirs.forEach(function(t) {
@@ -512,6 +524,6 @@ function runTest() {
 
 
 process.on('exit', function() {
-  assert.strictEqual(numtests, testsRun);
+  assert.strictEqual(2 * numtests, testsRun);
   assert.strictEqual(async_completed, async_expected);
 });
