@@ -1,13 +1,32 @@
+// Copyright Joyent, Inc. and other Node contributors.
+//
+// Permission is hereby granted, free of charge, to any person obtaining a
+// copy of this software and associated documentation files (the
+// "Software"), to deal in the Software without restriction, including
+// without limitation the rights to use, copy, modify, merge, publish,
+// distribute, sublicense, and/or sell copies of the Software, and to permit
+// persons to whom the Software is furnished to do so, subject to the
+// following conditions:
+//
+// The above copyright notice and this permission notice shall be included
+// in all copies or substantial portions of the Software.
+//
+// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
+// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
+// NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
+// DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
+// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
+// USE OR OTHER DEALINGS IN THE SOFTWARE.
+
 'use strict';
 const common = require('../common');
+if (!common.hasCrypto)
+  common.skip('node compiled without OpenSSL.');
+
 const assert = require('assert');
 const crypto = require('crypto');
 
-if (!common.hasCrypto) {
-  common.skip('node compiled without OpenSSL.');
-  return;
-}
-
 assert.throws(
   function() {
     crypto.getDiffieHellman('unknown-group');
@@ -35,7 +54,7 @@ assert.throws(
   'failed to throw the expected error.'
 );
 
-var hashes = {
+const hashes = {
   modp1: '630e9acd2cc63f7e80d8507624ba60ac0757201a',
   modp2: '18f7aa964484137f57bca64b21917a385b6a0b60',
   modp5: 'c0a8eec0c2c8a5ec2f9c26f9661eb339a010ec61',
@@ -47,24 +66,24 @@ var hashes = {
 };
 
 for (const name in hashes) {
-  var group = crypto.getDiffieHellman(name);
-  var private_key = group.getPrime('hex');
-  var hash1 = hashes[name];
-  var hash2 = crypto.createHash('sha1')
+  const group = crypto.getDiffieHellman(name);
+  const private_key = group.getPrime('hex');
+  const hash1 = hashes[name];
+  const hash2 = crypto.createHash('sha1')
                     .update(private_key.toUpperCase()).digest('hex');
-  assert.equal(hash1, hash2);
-  assert.equal(group.getGenerator('hex'), '02');
+  assert.strictEqual(hash1, hash2);
+  assert.strictEqual(group.getGenerator('hex'), '02');
 }
 
 for (const name in hashes) {
   // modp1 is 768 bits, FIPS requires >= 1024
   if (name === 'modp1' && common.hasFipsCrypto)
     continue;
-  var group1 = crypto.getDiffieHellman(name);
-  var group2 = crypto.getDiffieHellman(name);
+  const group1 = crypto.getDiffieHellman(name);
+  const group2 = crypto.getDiffieHellman(name);
   group1.generateKeys();
   group2.generateKeys();
-  var key1 = group1.computeSecret(group2.getPublicKey());
-  var key2 = group2.computeSecret(group1.getPublicKey());
+  const key1 = group1.computeSecret(group2.getPublicKey());
+  const key2 = group2.computeSecret(group1.getPublicKey());
   assert.deepStrictEqual(key1, key2);
 }
